using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Random;

namespace HeuristicLab.Problems.Instances.DataAnalysis {
  public class Feynman75 : FeynmanDescriptor {
    private readonly int testSamples;
    private readonly int trainingSamples;

    public Feynman75() : this((int) DateTime.Now.Ticks, 10000, 10000, null) { }

    public Feynman75(int seed) {
      Seed            = seed;
      trainingSamples = 10000;
      testSamples     = 10000;
      noiseRatio      = null;
    }

    public Feynman75(int seed, int trainingSamples, int testSamples, double? noiseRatio) {
      Seed                 = seed;
      this.trainingSamples = trainingSamples;
      this.testSamples     = testSamples;
      this.noiseRatio      = noiseRatio;
    }

    public override string Name {
      get {
        return string.Format("II.34.2a q*v/(2*pi*r) | {0}",
          noiseRatio == null ? "no noise" : string.Format(System.Globalization.CultureInfo.InvariantCulture, "noise={0:g}",noiseRatio));
      }
    }

    protected override string TargetVariable { get { return noiseRatio == null ? "I" : "I_noise"; } }

    protected override string[] VariableNames {
      get { return new[] {"q", "v", "r", noiseRatio == null ? "I" : "I_noise"}; }
    }

    protected override string[] AllowedInputVariables { get { return new[] {"q", "v", "r"}; } }

    public int Seed { get; private set; }

    protected override int TrainingPartitionStart { get { return 0; } }
    protected override int TrainingPartitionEnd { get { return trainingSamples; } }
    protected override int TestPartitionStart { get { return trainingSamples; } }
    protected override int TestPartitionEnd { get { return trainingSamples + testSamples; } }

    protected override List<List<double>> GenerateValues() {
      var rand = new MersenneTwister((uint) Seed);

      var data = new List<List<double>>();
      var q    = ValueGenerator.GenerateUniformDistributedValues(rand.Next(), TestPartitionEnd, 1, 5).ToList();
      var v    = ValueGenerator.GenerateUniformDistributedValues(rand.Next(), TestPartitionEnd, 1, 5).ToList();
      var r    = ValueGenerator.GenerateUniformDistributedValues(rand.Next(), TestPartitionEnd, 1, 5).ToList();

      var I = new List<double>();

      data.Add(q);
      data.Add(v);
      data.Add(r);
      data.Add(I);

      for (var i = 0; i < q.Count; i++) {
        var res = q[i] * v[i] / (2 * Math.PI * r[i]);
        I.Add(res);
      }

      if (noiseRatio != null) {
        var I_noise     = new List<double>();
        var sigma_noise = (double) Math.Sqrt(noiseRatio.Value) * I.StandardDeviationPop();
        I_noise.AddRange(I.Select(md => md + NormalDistributedRandom.NextDouble(rand, 0, sigma_noise)));
        data.Remove(I);
        data.Add(I_noise);
      }

      return data;
    }
  }
}